import { cpus } from 'os';
// eslint-disable-next-line import/no-unresolved
import { Worker } from 'worker_threads';
const numCpus = cpus().length;
export class Ework {
    constructor(worker) {
        const workerString = worker.toString();
        const workerCode = makeWorkerCode(workerString);
        this.totalWorkers = Math.max(numCpus - 1, 2);
        this.freeWorkers = this.totalWorkers;
        this.workers = [];
        for (let i = 0; i < this.totalWorkers; i++) {
            const worker = new Worker(workerCode, { eval: true });
            const workerObj = {
                worker,
                isWorking: false,
                job: null,
            };
            worker.on('message', (message) => {
                if (message.type === 'result') {
                    const job = workerObj.job;
                    if (job === null) {
                        throw new Error('UNREACHABLE');
                    }
                    if (job.id !== message.id) {
                        throw new Error('UNREACHABLE');
                    }
                    if (message.status === 'success') {
                        job.resolve(message.result);
                    }
                    else {
                        job.reject(new Error(message.error));
                    }
                    workerObj.job = null;
                    workerObj.isWorking = false;
                    this.freeWorkers++;
                    this.run();
                }
            });
            this.workers.push(workerObj);
        }
        this.queue = [];
        this.nextWorkId = 0;
    }
    async execute(value) {
        return this.enqueue(value);
    }
    async terminate() {
        await Promise.all(this.workers.map((w) => new Promise((resolve) => {
            w.worker.terminate(() => resolve());
            if (w.job !== null) {
                w.job.reject(new Error('worker terminated'));
            }
        })));
        this.freeWorkers = 0;
        this.queue = [];
        this.workers = [];
    }
    async map(values) {
        return Promise.all(values.map((value) => this.enqueue(value)));
    }
    enqueue(value) {
        let resolve;
        let reject;
        const promise = new Promise((res, rej) => {
            resolve = res;
            reject = rej;
        });
        const workload = {
            id: this.nextWorkId++,
            value,
            // @ts-ignore
            resolve,
            // @ts-ignore
            reject,
        };
        this.queue.push(workload);
        this.run();
        return promise;
    }
    run() {
        while (this.freeWorkers > 0 && this.queue.length > 0) {
            const nextJob = this.queue.shift();
            if (nextJob === undefined) {
                throw new Error('UNREACHABLE');
            }
            const worker = this.getFreeWorker();
            worker.worker.postMessage({
                id: nextJob.id,
                type: 'work',
                value: nextJob.value,
            });
            worker.job = nextJob;
            worker.isWorking = true;
            this.freeWorkers--;
        }
    }
    getFreeWorker() {
        const worker = this.workers.find((w) => !w.isWorking);
        if (worker === undefined) {
            throw new Error('UNREACHABLE');
        }
        return worker;
    }
}
function makeWorkerCode(workerString) {
    return `'use strict';
const { parentPort } = require('worker_threads');

const workerFunction = ${workerString};

parentPort.on('message', async (message) => {
  if (message.type === 'work') {
    try {
      const result = await workerFunction(message.value);
      parentPort.postMessage({
        id: message.id,
        type: 'result',
        status: 'success',
        result
      });
    } catch (error) {
      let errorToTransfer = error;
      if (error) {
        if (error.stack) {
          errorToTransfer = error.stack;
        } else if (error.message) {
          errorToTransfer = error.message;
        }
      }
      try {
        parentPort.postMessage({
          id: message.id,
          type: 'result',
          status: 'error',
          error: errorToTransfer
        });
      } catch (e) {
        parentPort.postMessage({
          id: message.id,
          type: 'result',
          status: 'error',
          error: 'Work failed but error could not be transferred: ' + e.message
        });
      }
    }
  }
});
  `;
}
//# sourceMappingURL=Ework.js.map